"""
Contains field validation functions
"""

import logging
import re

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

email_regex = re.compile('^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z.]+$', flags=re.IGNORECASE)
telephone_regex = re.compile('^\+?\d+(\s*-?\s*\d+)*$')
postcode_regex = re.compile('^[A-Za-z0-9 -]{1,10}$')
url_regex = re.compile('^https?://[^.]+\.[^.]+')
card_number_regex = re.compile('^\d{16}$')
security_code_regex = re.compile('^\d{3}$')
royal_mail_tracking_number_regex = re.compile('^[A-Z0-9]{11}GB$')
url_safe_regex = re.compile('^[A-Za-z0-9_-]+$')


def luhn_check(card_number):
    """ checks to make sure that the card passes a luhn mod-10 checksum """
    sum = 0
    num_digits = len(card_number)
    oddeven = num_digits & 1

    for count in range(0, num_digits):
        digit = int(card_number[count])

        if not ((count & 1) ^ oddeven):
            digit *= 2
        if digit > 9:
            digit -= 9

        sum += digit

    return (sum % 10) == 0


def validate_email_address(email_address):
    return email_regex.match(email_address)


def validate_url(url):
    return url_regex.match(url)


def validate_telephone_number(telephone_number):
    return telephone_regex.match(telephone_number) and len(telephone_number) >= 8


def validate_postcode(postcode):
    return postcode_regex.match(postcode)


def validate_card_number(card_number):
    return card_number_regex.match(card_number)


def validate_security_code(security_code):
    return security_code_regex.match(security_code)


def validate_royal_mail_tracking_number(tracking_number):
    return royal_mail_tracking_number_regex.match(tracking_number)


def validate_url_safe(value):
    return url_safe_regex.match(value)




"""
This contains the SessionVar classes - use these to add variables to your SessionData classes
"""

import logging
import decimal
from collections import Mapping

from .framework import SessionVar as _Var

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class SessionInt(_Var):
    def to_json_val(self, val):
        return int(val)


class SessionStr(_Var):
    def to_json_val(self, val):
        return str(val)


class SessionBool(_Var):
    def to_json_val(self, val):
        return bool(val)


class SessionDecimal(_Var):
    def to_json_val(self, val):
        if not isinstance(val, (decimal.Decimal, int)):
            raise ValueError('SessionDecimal field must be set to a Decimal (or int) value')

        return str(val)

    def from_json_val(self, json_val):
        return decimal.Decimal(json_val)


class SessionDict(_Var):
    def __init__(self, key, template_dict, default_value=None):
        """
        Stores a dictionary in the session

        :param key: The key to store this under
        :param template_dict: Dictionary, mapping dictionary keys onto SessionVars, defining the
                              type of data that should be mapped under each key. The SessionVars'
                              key attributes will be ignored - you can just pass in key=None
        """
        super().__init__(key, default_value=default_value)

        self.template_dict = template_dict

    def to_json_val(self, val):
        # Make sure this is a dict-like object
        if not isinstance(val, Mapping):
            raise ValueError('Non Mapping value "{}" cannot be stored in dict field'
                             .format(val))

        # Check for invalid keys
        for key in val:
            if key not in self.template_dict:
                raise ValueError('Invalid key "{}" in dict to be stored in session: {}'
                                 .format(key, val))

        # Build the dict to be stored
        json_dict = {}
        for key in self.template_dict:
            if key not in val:
                raise ValueError('Key "{}" not found in dict to be stored in session: {}'
                                 .format(key, val))

            session_var = self.template_dict[key]
            json_dict[key] = session_var.to_json_val(val[key])

        return json_dict

    def from_json_val(self, json_val):
        # Check for invalid keys
        for key in json_val:
            if key not in self.template_dict:
                raise ValueError('Invalid key "{}" in dict stored in session: {}'
                                 .format(key, json_val))

        # Build the dict to be stored
        out_dict = {}
        for key in self.template_dict:
            if key not in json_val:
                raise ValueError('Key "{}" not found in dict stored in session: {}'
                                 .format(key, json_val))

            session_var = self.template_dict[key]
            out_dict[key] = session_var.from_json_val(json_val[key])

        return out_dict


class SessionList(_Var):
    def __init__(self, key, item_type, default_value=None):
        """
        :param item_type: SessionVar defining the type of each item in the list.  The key attribute
        of item_type will be ignored - you can just pass in None
        """
        if isinstance(item_type, type):
            raise ValueError('item_type needs to be an instance of a SessionVar and not a class')
        
        super().__init__(key, default_value=default_value)
        
        self.item_type = item_type

    def to_json_val(self, val):
        if not isinstance(val, list):
            raise ValueError('Non list value "{}" cannot be stored in list session variable'
                             .format(val))
        json_list = []
        for i in val:
            json_list.append(self.item_type.to_json_val(i))

        return json_list

    def from_json_val(self, json_val):
        out_list = []
        for i in json_val:
            out_list.append(self.item_type.from_json_val(i))

        return out_list

        
        



"""
Contains code for generating random placeholder text
"""

import logging
import random
import os

import pymarkovchain

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

MAX_ATTEMPTS = 30
PUNCTUATION = '...........?!!'

ADJECTIVES = [
    'able', 'abundant', 'accepting', 'accommodating', 'active', 'addictive', 'adequate',
    'aggressive', 'amazing', 'ambiguousbarren', 'amiable', 'amicable', 'amusing',
    'antagonistic', 'anxious', 'apathetic', 'aquatic', 'arrogant', 'articulate',
    'artistic', 'attentive', 'attractive', 'authoritative', 'awesome',
    
    'benevolent', 'biodegradable', 'blase', 'bold', 'bonding', 'boorish', 'bountiful',
    'braggart', 'brave', 'brilliant', 'buoyant', 'busy', 'buzzing',
    
    'callow', 'captious', 'caring', 'celestial', 'charming', 'chaste', 'cheating',
    'cheerful', 'churlish', 'civil', 'clean', 'clever', 'coastal', 'cold', 'colossal',
    'combustible', 'comfortable', 'commercial', 'communicative', 'compact',
    'competitive', 'compulsive', 'confident', 'conflicted', 'congenial', 'conscientious',
    'conservative', 'considerate', 'conspicuous', 'contemptible', 'contiguous',
    'cooperative', 'cordial', 'courageous', 'courteous', 'covetous', 'creative',
    'critical', 'critical', 'crucial', 'crude', 'culpable', 'curious', 'current', 'curt',
    'cynical', 'cool',
    
    'decent', 'decorous', 'defensive', 'deferential', 'deft', 'dejected', 'delightful',
    'demeaning', 'demise', 'dependable', 'deplorable', 'depressed', 'destructive',
    'devious', 'devoted', 'dictatorial', 'diligent', 'diminutive', 'diplomatic',
    'discreet', 'disdainful', 'dishonest', 'dishonorable', 'disposable', 'disrespectful',
    'distracted', 'docile', 'downcast', 'dynamic',
    
    'earnest', 'earthy', 'ecological', 'efficient', 'egotistical', 'electrifying', 'elitist',
    'empathetic', 'endangered', 'endemic', 'energetic', 'enigmatic', 'enthusiastic',
    'esteemed', 'estimable', 'ethical', 'euphoric', 'evergreen', 'exclusive',
    'expectant', 'explosive', 'exquisite', 'extravagant', 'extrovert', 'exuberant',
    
    'fair', 'faithful', 'fallow', 'false', 'famous', 'fancy', 'ferocious', 'fertile',
    'fervent', 'fervid', 'fibrous', 'fierce', 'flexible', 'focused', 'forgiving',
    'forlorn', 'frail', 'foolish',
    
    'generous', 'genial', 'genteel', 'gentle', 'genuine', 'gifted', 'gigantic', 'glib',
    'gloomy', 'good', 'gorgeous', 'grace', 'gracious', 'grand', 'grateful',
    'green', 'grouchy', 'guilty', 'guilty', 'gusty',
    
    'happy', 'hard-hearted', 'healing', 'heedless', 'helpful', 'heroic', 'honest',
    'honorable', 'hopeful', 'hostile', 'humane', 'humble', 'humorous', 'hygienic',
    'hysterical',
    
    'idealistic', 'idolize', 'ignoble', 'ignorant', 'ill-tempered', 'impartial', 'impolite',
    'improper', 'imprudent', 'impudent', 'indecent', 'indecorous', 'indifferent',
    'indigenous', 'industrious', 'ingenuous', 'innocent', 'innovative', 'insightful',
    'insolent', 'inspirational', 'instructive', 'insulting', 'intense', 'intense',
    'intense', 'intolerant', 'introvert', 'intuitive', 'inventive', 'investigative',
    'irascible', 'irresponsible',
    
    'jaundiced', 'jealous', 'jealous', 'jocular', 'jolly', 'jovial', 'joyful', 'jubilant',
    'just', 'juvenile',
    
    'keen', 'kind', 'kindred', 'knowledgeable',
    
    'liberal', 'listener', 'loathsome', 'loving', 'loyal',
    
    'magical', 'magnificent', 'malevolent', 'malicious', 'mammoth',
    'manipulative', 'marine', 'mastery', 'meddling', 'meritorious', 'meticulous',
    'migratory', 'minuscule', 'miserable', 'mistrustful', 'modest', 'moral',
    'mysterious',
    
    'naive', 'nascent', 'native', 'natural', 'natural', 'needy', 'nefarious', 'negative',
    'neglected', 'neglectful', 'negligent', 'nice', 'noble', 'notorious', 'nonsensical',
    
    'obedient', 'observant', 'open', 'open-minded', 'opinionated', 'oppressive', 'orderly',
    'oriented', 'original', 'outrageous', 'outspoken', 'over-the-top',
    
    'parasitic', 'partial', 'passionate', 'patient', 'perceptive', 'personable', 'personal',
    'petulant', 'pleasant', 'poise', 'polite', 'pollutant', 'popular', 'popular',
    'powerful', 'prejudicial', 'preposterous', 'pretentious', 'prideful', 'principled',
    'pristine', 'prompt', 'proper', 'punctual', 'purposeful',
    
    'quaint', 'quarrelsome', 'quick', 'quiet', 'quirky',
    
    'radioactive', 'rancorous', 'rational', 'reasonable', 'reckless', 'refined',
    'reflective', 'reliant', 'remarkable', 'remorseful', 'renewable', 'reproductive',
    'repugnant', 'resilient', 'resilient', 'resolute', 'resourceful', 'respectful',
    'responsible', 'responsive', 'restorative', 'reverent', 'rotting', 'rude',
    'ruthless', 'religious',
    
    'sadness', 'safe', 'scornful', 'scrumptious', 'selfish', 'sensible', 'sensitive',
    'sharing', 'simple', 'sober', 'solar', 'solemn', 'solitary', 'soluble', 'sour',
    'spatial', 'special', 'splendid', 'splendid', 'staunch', 'staunch', 'stern',
    'stunning', 'successful', 'sullen', 'superb', 'superior', 'supportive', 'surly',
    'suspicious', 'sweet', 'sympathetic', 'single',
    
    'tactful', 'tainted', 'temperate', 'temperate', 'tenacious', 'terrific', 'testy',
    'thoughtful', 'thoughtless', 'tolerant', 'towering', 'toxic', 'treacherous',
    'tropical', 'trustworthy', 'truthful',
    
    'ultimate', 'ultimate', 'uncivil', 'uncouth', 'undeveloped', 'unethical', 'unfair',
    'unique', 'unique', 'united', 'unity', 'unmannerly', 'unrefined', 'unsavory',
    'unworthy', 'uplifting', 'upright', 'uprooted', 'upstanding',
    
    'valiant', 'veracious', 'versatile', 'vicious', 'vigilant', 'vigorous', 'vile',
    'villainous', 'virtuous', 'visible', 'vivacious', 'vocal', 'volatile',
    'volunteering', 'vulnerable',
    
    'warm', 'wary', 'waspish', 'watchful', 'welcoming', 'wicked', 'wild', 'willingness',
    'winning', 'winsome', 'wise', 'wishy-washy', 'wistful', 'witty', 'woeful',
    'wonderful', 'worldwide', 'worrier', 'worthwhile', 'worthy',
    
    'yearning', 'yielding', 'youthful',
    
    'zany', 'zealous', 'zero-tolerant'
]

NOUNS = [
    'addition', 'adjustment', 'aftermath', 'afternoon', 'agreement', 'airport', 'alarm',
    'albino', 'ambiguity', 'angle', 'arch', 'astonishment',
    
    'babies', 'bait', 'balance', 'ball-bearing', 'basket', 'bat', 'bath', 'bell', 'bike',
    'birds', 'birthday', 'bit', 'boat', 'bomb', 'bone', 'books', 'brass', 'breath',
    'brick', 'brother', 'burst', 'business', 'butter',
    
    'calculator', 'camera', 'cannon', 'car', 'care', 'carpenter', 'carriage', 'cart',
    'celery', 'cent', 'chance', 'cheese', 'cherries', 'cherry', 'chicken', 'class',
    'cobweb', 'color', 'company', 'connection', 'control', 'copper', 'cows', 'crow',
    'cup', 'corncob',
    
    'danger', 'death', 'desk', 'digestion', 'discussion', 'dog', 'dogs', 'drain', 'dress',
    'drop', 'duck', 'ducks',
    
    'ear', 'elbow', 'elephant', 'empathy', 'exchange', 'expert', 'explosion', 'eye',
    
    'face', 'fall', 'finger', 'finger', 'fish', 'flesh', 'flower', 'food', 'fork', 'fowl',
    'fractal',
    
    'giants', 'gin', 'girl', 'glass', 'glove', 'governor', 'grain', 'grandfather', 'grin', 'grip',
    'ground', 'guide', 'gun', 'gravity',
    
    'hair', 'hammer', 'harmony', 'hate', 'head', 'hilarity', 'holiday', 'honey', 'horse',
    'humor',
    
    'impulse', 'incompetence', 'increase', 'increase', 'industry', 'instrument', 'interest',
    
    'jelly', 'jelly bean', 'judge', 'jump', 'jungle',
    
    'kelp', 'kettle', 'kitty', 'knee',
    
    'lady', 'ladybug', 'lake', 'leg', 'library', 'locket', 'loss', 'lumber', 'lunchroom',
    
    'magic', 'man', 'market', 'measure', 'meatloaf', 'meeting', 'memory', 'middle', 'mind', 'mine',
    'monkey', 'monster', 'morning', 'mother', 'move',
    
    'needle', 'newt', 'night', 'noise', 'nothing', 'number', 'nut',
    
    'oatmeal', 'observation', 'offal', 'office', 'orange',
    
    'page', 'part', 'passenger', 'person', 'pickle', 'picture', 'pies', 'plate', 'play',
    'playground', 'pleasure', 'plot', 'position', 'pot', 'powder', 'power', 'proton',
    'pump',
    
    'queen', 'quiet', 'quill', 'quince', 'quiver',
    
    'rabbit', 'rail', 'railway', 'rat', 'rate', 'reality', 'reason', 'receipt',
    'request', 'rhythm', 'rice', 'riddle', 'ridicule', 'rifle', 'road', 'roof', 'room', 'rule',
    'run',
    
    'sand', 'sandwich', 'sail', 'scene', 'scissors', 'seashore', 'secretary', 'servant', 'sheep',
    'sheet', 'shelf', 'show', 'side', 'sink', 'skate', 'sky', 'sleep', 'smash', 'smile',
    'snails', 'snake', 'sock', 'sound', 'space', 'spot', 'spring', 'stamp', 'star',
    'start', 'stocking', 'stone', 'street', 'suggestion', 'sweater',
    
    'taste', 'temper', 'territory', 'theory', 'thing', 'thunder', 'ticket', 'toad', 'toes',
    'tooth', 'toothbrush', 'toothpaste', 'top', 'town', 'toys', 'transport', 'truck',
    'turkey', 'turnip', 'tyrant',
    
    'underwear', 'utopia',
    
    'van', 'vein', 'vernacular', 'voice',
    
    'wall', 'war', 'water', 'wave', 'wax', 'wealth', 'wire', 'wish', 'wind', 'woman', 'wool',
    'wound', 'wren', 'wrist',
    
    'xylophone',
    
    'yak', 'yard', 'yarn', 'yoke',
    
    'zen'
]

QUALIFIERS = [
    'rather', 'quite', 'very', 'exceedingly', 'pretty', 'almost', 'not quite',
    'absolutely', 'mostly', 'almost', 'most definitely', 'disturbingly', 'ridiculously',
    'spectacularly', 'slightly', 'nearly', 'annoyingly', 'ambiguously', 'possibly',
    'disproportionately', 'seemingly'
]


class TextGenerator(object):
    def __init__(self, filename=None, escape_quotes=False):
        """
        :param filename: Path to a markov chain data file (generated using markov_tool)
        """
        if filename is None:
            path = os.path.dirname(os.path.realpath(__file__))
            filename = os.path.join(path, 'test_data', 'markov.dat')
        
        log.info('Loading Markov Chains from "{}"'.format(filename))

        self.markov = pymarkovchain.MarkovChain(filename)
        self.escape_quotes = escape_quotes
        self.nouns = NOUNS
        self.adjectives = ADJECTIVES
        self.qualifiers = QUALIFIERS
    
    def _process_text(self, text):
        if self.escape_quotes:
            text = text.replace('"', '&quot;')
        return text

    def generate_text(self, min=None, max=None):
        for i in range(MAX_ATTEMPTS):
            result = self.markov.generateString()
            l = len(result)
            if min and l < min:
                continue
            if max and l > max:
                continue

            return self._process_text(result)
        
        # We couldn't get one of the right length
        if max > 5:
            result = result[:max - 3] + '...'
        
        return self._process_text(result)

    def generate_sentence(self, min=None, max=None, punctuation=True):
        end = random.choice(PUNCTUATION) if punctuation else ''
        return self.generate_text(min=min, max=max) + end

    def generate_sentences(self, min_sentences=1, max_sentences=6, min_sentence_length=30, max_sentence_length=200):
        sentences = []
        num_sentences = random.randint(min_sentences, max_sentences)
        for i in range(num_sentences):
            sentences.append(self.generate_sentence(min=min_sentence_length, max=max_sentence_length))

        return ' '.join(sentences)

    def generate_html_paragraph(self, min_sentences=1, max_sentences=6, min_sentence_length=30, max_sentence_length=200):
        return '<p>{}</p>'.format(self.generate_sentences(min_sentences, max_sentences,
                                                          min_sentence_length,
                                                          max_sentence_length))
   
    def generate_html_paragraphs(self, min_paragraphs=2, max_paragraphs=5, min_sentences=1, max_sentences=7, min_sentence_length=10, max_sentence_length=130):
        paragraphs = []
        num_paragraphs = random.randint(min_paragraphs, max_paragraphs)
        for i in range(num_paragraphs):
            paragraphs.append(self.generate_html_paragraph(min_sentences, max_sentences,
                                                           min_sentence_length, max_sentence_length))

        return '\n'.join(paragraphs)

    def generate_html_list(self, min_items=3, max_items=7, min_item_length=20, max_item_length=100):
        ul = '<ul>'
        num_list_items = random.randint(min_items, max_items)
        for i in range(num_list_items):
            ul += '<li>{}</li>'.format(self.generate_text(min=min_item_length, max=max_item_length))
        ul += '</ul>'

        return ul

    def generate_adjective_noun(self, qualifier_chance=0.3, the_chance=0.1, adjective_chance=0.9, the_double_noun_chance=0.08):
        if random.random() < the_double_noun_chance:
            return 'the {} {}'.format(random.choice(self.nouns), random.choice(self.nouns))

        parts = []
        if random.random() < the_chance:
            parts.append('the')
        if random.random() < adjective_chance:
            if random.random() < qualifier_chance:
                parts.append(random.choice(self.qualifiers))
            parts.append(random.choice(self.adjectives))
        parts.append(random.choice(self.nouns))

        return ' '.join(parts)

    def generate_adjective(self):
        return random.choice(self.adjectives)

    def generate_noun(self):
        return random.choice(self.nouns)




"""
Generates a string on startup that uniquely identifies this host
"""

import logging
import random
import uuid

from littlefish import timetool

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

identifier = None
host_uuid = None


def init():
    global identifier, host_uuid

    if identifier:
        raise Exception('hostidentifier.init called multiple times!')

    # In production we may have multiple instances of the application on the same server
    # so we need a way to identify each instance in the log file
    timestamp = int((timetool.unix_time() * 3000) % 2 ** 12)
    r = ''.join([random.choice('0123456789ABCDEF') for i in range(3)])
    identifier = '{:03X}-{}'.format(timestamp, r)

    print('This host is identified by {}'.format(identifier))

    host_uuid = uuid.uuid4()

    print('Host UUID: {}'.format(host_uuid))



"""
"Scanner" to scan through front end content and:

- Generate a sitemap
- Check for missing meta tags
- Check for missing alt tags

Example Usage:

    BLUEPRINTS = ['main']
    SKIP = ['main.super_secret_test_stuff', 'main.site_map']

    scanner = ContentScanner(app, blueprints=BLUEPRINTS, skip=SKIP, scheme='https')
    
    # Add some argument functions for dynamic content
    @scanner.args_function('main.view_product_details')
    def view_product_details_args():
        products = Product.query.all()
        for product in products:
            yield {'url_name': product.url_name}

    @scanner.args_function('main.view_post')
    def view_post_args():
        news_posts = NewsPost.query.all()
        for post in news_posts:
            yield {'nice_name': post.nice_name}
    
    # Get the results
    results = scanner.scan(True, True, True, True, True)
"""

import logging
import re
import datetime

from werkzeug.wrappers import BaseResponse
from werkzeug.routing import parse_rule, BuildError
from flask import url_for
from bs4 import BeautifulSoup
import titlecase

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class ViewRaisedException(Exception):
    """
    The view function raised an unhandled exception when it was run
    """
    pass


class ConfigurationError(Exception):
    """
    There is an error in the scanner configuration
    """
    pass


class InvalidViewArguments(Exception):
    """
    The arguments passed into the url generator were incorrect
    """
    pass


class HtmlParsingError(Exception):
    """
    There was a problem parsing the HTML when inspecting the content
    """
    pass


class ScannerResult:
    def __init__(self, endpoint, args, url=None, status_code=None, mime_type=None, title=None,
                 meta_description=None, missing_alt_tags=[], last_modified=None):
        self.endpoint = endpoint
        self.args = args
        self.url = url
        self.status_code = status_code
        self.mime_type = mime_type
        self.title = title
        self.meta_description = meta_description
        self.missing_alt_tags = missing_alt_tags
        if last_modified:
            self.last_modified = last_modified
        else:
            self.last_modified = datetime.datetime.utcnow()
    
    @property
    def is_html(self):
        return 'html' in self.mime_type

    @property
    def needs_meta_description(self):
        if not self.is_html:
            return False
        
        if self.status_code != 200:
            return False

        return self.meta_description is None or len(self.meta_description) < 10


class SiteMapEntry(object):
    def __init__(self, url, modified, change_freq, priority, title):
        self.url = url
        self.modified = modified
        self.change_freq = change_freq
        self.priority = priority
        self.title = title


class ContentScanner:
    def __init__(self, app, blueprints=None, skip=[], scheme='http', server_name=None,
                 ignore_images_containing=[]):
        """
        :param app: The flask app
        :param blueprints: List of blueprints to scan, or None to scan all of them
        :param skip: List of endpoints to skip.  The list can contain regular expressions and strings
        :param scheme: http or https - for url generation
        :param server_name: The server name used for url generation.  If this is not specified we will attempt
                            to extract this from the SERVER_NAME variable in app.config
        :param ignore_images_containing: When scanning images, images with sources which contain
                                         any of the strings in this list will be ignored
        """
        self.app = app
        self.blueprints = blueprints
        self.skip_strings = [s for s in skip if isinstance(s, str)]
        self.skip_regexps = [r for r in skip if hasattr(r, 'search')]
        assert len(self.skip_strings) + len(self.skip_regexps) == len(skip), \
            'Skip list must only contain strings and regular expressions'
        self.scheme = scheme
        # Maps endpoint -> function to get arguments
        self.endpoint_argument_functions = {}
        if server_name is None:
            if not app.config.get('SERVER_NAME'):
                raise ValueError('No server name passed in and no SERVER_NAME specified in flask config')

            self.server_name = app.config['SERVER_NAME']
        else:
            self.server_name = server_name
        self.ignore_images_containing = ignore_images_containing

    def args_function(self, endpoint):
        """
        Decorator.

        This function must yield dicts with the kwargs that need to be passed into the view function

        :param endpoint: The endpoint that we are defining args for
        """
        if endpoint in self.endpoint_argument_functions:
            raise ConfigurationError('Added multiple arg functions for enpoint: {}'.format(endpoint))

        def decorator(f):
            self.endpoint_argument_functions[endpoint] = f
            return f

        return decorator

    def scan(self, generate_urls=False, inspect_content=False, inspect_meta_description=False,
             inspect_alt_tags=False, inspect_title=False):
        """
        Scan through the content.

        :param generate_urls: Whether or not to include urls in the result set
        :param inspect_content: If True, this will scan through for mime types and status codes. If any
                                of the other inspect_ parameters are set to True this will happen
                                automatically
        :param inspect_meta_description: Whether or not to collect meta description tags
        :param inspect_alt_tags: Whether or not to compile a list of img tags with no alt tags for each
                                 endpoint
        :param inspect_title: Whether or not to extract titles from the html
        """

        if not inspect_content:
            inspect_content = inspect_meta_description or inspect_alt_tags or inspect_title
        
        results = []

        for rule in self.app.url_map.iter_rules():
            endpoint_parts = rule.endpoint.split('.')
            blueprint_name = endpoint_parts[0] if len(endpoint_parts) > 1 else None

            if self.blueprints is not None and blueprint_name not in self.blueprints:
                log.debug('Skipping rule {} due to blueprint ({})'.format(
                    rule, blueprint_name
                ))
                continue
            
            skip = False

            if rule.endpoint in self.skip_strings:
                skip = True
            else:
                for regexp in self.skip_regexps:
                    if regexp.search(rule.endpoint):
                        skip = True
                        break

            if skip:
                log.debug('Skipping rule {}'.format(rule))
                continue

            log.debug('[{}] {} {} {}'.format(
                blueprint_name, rule.endpoint, rule.methods, rule
            ))

            results += self.scan_rule(rule, generate_urls, inspect_content, inspect_meta_description,
                                      inspect_alt_tags, inspect_title)

        return results

    def scan_rule(self, rule, generate_urls, inspect_content, inspect_meta_description, inspect_alt_tags,
                  inspect_title):

        inspect_html = inspect_meta_description or inspect_alt_tags or inspect_title

        view_function = self.app.view_functions[rule.endpoint]
        
        # Count up the number of arguments
        num_args = 0
        for converter, arguments, variable in parse_rule(str(rule)):
            if converter:
                num_args += 1

        if num_args > 0:
            # The function has args!
            args_function = self.endpoint_argument_functions.get(rule.endpoint)
            if not args_function:
                raise ConfigurationError('No args function for endpoint "{}". You need to use the '
                                         'args_function decorator to add a function that yields all of the '
                                         'combinations of arguments for this function'.format(rule.endpoint))
        else:
            def args_function():
                return [{}]
        
        results = []
        
        base_url = '{}://{}'.format(self.scheme, self.server_name)

        for kwargs in args_function():
            log.debug(' > args: {}'.format(kwargs))
            with self.app.test_request_context(base_url=base_url):
                result = ScannerResult(rule.endpoint, kwargs)
                
                if generate_urls:
                    try:
                        result.url = url_for(rule.endpoint, _external=True, _scheme=self.scheme, **kwargs)
                    except BuildError as e:
                        raise InvalidViewArguments('The arguments {} are not valid for endpoint {}'
                                                   .format(kwargs, rule.endpoint)) from e

                if inspect_content:
                    try:
                        response = view_function(**kwargs)
                    except Exception as e:
                        if kwargs:
                            error_message = 'Endpoint {} with args {} raised an exception'.format(
                                rule.endpoint, kwargs
                            )
                        else:
                            error_message = 'Endpoint {} raised an exception'.format(
                                rule.endpoint
                            )

                        raise ViewRaisedException(error_message) from e

                    if isinstance(response, str):
                        response_text = response
                        status_code = 200
                        mimetype = 'text/html'
                    elif isinstance(response, BaseResponse):
                        status_code = response.status_code
                        mimetype = response.mimetype
                        response_text = ''
                        if mimetype.startswith('text/'):
                            for response_part in response.response:
                                response_text += response_part.decode('utf-8')
                        if response.last_modified:
                            result.last_modified = response.last_modified
                    elif isinstance(response, tuple):
                        if len(response) != 2:
                            raise ValueError('Unhandled responstype: tuple with length {}'
                                             .format(len(response)))
                        response_text, status_code = response
                        if not isinstance(response_text, str):
                            raise ValueError('Unhandled response_type: tuple with non str as first item')
                        mimetype = 'text/html'
                    else:
                        raise ValueError('Unhandled response type: {}'.format(type(response)))
                    
                    truncated_response_text = response_text.replace('\n', '')[:30]
                    log.debug('{} ({}) {}...'.format(
                        status_code, mimetype, truncated_response_text
                    ))
                    result.status_code = status_code
                    result.mime_type = mimetype

                    if inspect_html and 'html' in result.mime_type:
                        soup = BeautifulSoup(response_text, 'html.parser')

                        if inspect_meta_description:
                            head = soup.find('head')
                            if head:
                                meta_desc_tags = head.find_all('meta', attrs={"name": re.compile(r"^\s*description\s*$", re.I)})
                                if len(meta_desc_tags) > 1:
                                    all_tags = '; '.join([str(t) for t in meta_desc_tags])
                                    raise HtmlParsingError(
                                        'Found multiple description tags for endpoint {}: {}'
                                        .format(rule.endpoint, all_tags)
                                    )
                                elif len(meta_desc_tags) == 1:
                                    result.meta_description = meta_desc_tags[0]['content']

                        if inspect_alt_tags:
                            missing_alt_tags = []
                            all_images = soup.find_all('img')
                            for image in all_images:
                                # Sometimes we may need to skip some images
                                if self.ignore_images_containing:
                                    src = image.get('src')
                                    skip = False
                                    for image_path_search in self.ignore_images_containing:
                                        if src and image_path_search in src:
                                            skip = True
                                            break
                                    if skip:
                                        continue

                                alt = image.get('alt')
                                if alt is None or not alt.strip():
                                    missing_alt_tags.append(str(image))
                            result.missing_alt_tags = missing_alt_tags

                        if inspect_title:
                            head = soup.find('head')
                            if head:
                                title_tag = head.find('title')
                                if title_tag:
                                    result.title = title_tag.text.strip()

                results.append(result)

        return results

    def scan_for_sitemap(self, priority_map={}, change_freq='daily', default_priority='0.5',
                         inspect_content=False, inspect_title=False):
        # Scan through the pages and return a list of sitemap entries
        results = self.scan(generate_urls=True, inspect_content=inspect_content, inspect_title=inspect_title)
        sitemap = []

        for result in results:
            if inspect_content:
                if result.status_code != 200:
                    continue

                if not result.is_html:
                    continue

            # priority_map maps endpoint -> priority
            priority = priority_map.get(result.endpoint, default_priority)
            title = result.title
            if not title:
                # Make one up!
                title = titlecase.titlecase(result.endpoint.split('.')[1].replace('_', ' '))

            sitemap.append(SiteMapEntry(result.url, result.last_modified, change_freq, priority, title))

        return sitemap




"""
This module uses endesive to sign PDF files and provides
some classes to manage this and to help render the signature
into the body of the PDF file (similar for portable signer)
"""

import logging
import datetime

import OpenSSL.crypto
import endesive.pdf
import pytz

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class Signature:
    """
    Represents a signature on a PDF
    """
    def __init__(self, contact, issuer, serial_number, location, reason, flags=3,
                 timezone='Europe/London'):
        """
        :param contact: The person signing the PDF eg 'Bob the Builder'
        :param issuer: The certificate issuer eg
                       'C=GB, ST=Greater Manchester, L=Salford, O=Sectigo Limited,
                        CN=Sectigo RSA Client Authentication and Secure Email CA'
        :param serial_number: Certificate serial number
        :param location: Location of signing eg 'Some Place, Acton, London'
        :param reason: Why are you signing this? eg 'Customer Contract'
        :param flags: I have no idea what this is but it works if you set it to 3!
        :param timezone: You can override the timezone with this
        """
        self.flags = flags
        self.contact = contact
        self.issuer = issuer
        self.serial_number = serial_number
        self.location = location
        self.signing_date = datetime.datetime.utcnow()
        self.reason = reason

        local_timezone = pytz.timezone(timezone)
        self.timezone = local_timezone.tzname(self.signing_date)

    def get_signature_args(self):
        """
        Get the arguments in the weird format needed for signing
        """
        timestamp = self.signing_date.strftime('%Y%m%d%H%M%S+00\'00\'')

        return {
            b'sigflags': 3,
            b'contact': self.contact.encode(),
            b'location': self.location.encode(),
            b'signingdate': timestamp.encode(),
            b'reason': self.reason.encode()
        }


class Signer:
    """
    A Signer object is used to sign PDF files.  Create one and use it to sign stuff!

    Create a Signer, then pass in Signer.signature to your PDF generator if you want
    to render the signature.  Then get the PDF file (bytes) and pass it into
    get_signed_pdf(...) to sign it.
    """
    def __init__(self, pfx_path, password, location, reason):
        """
        :param pfx_path: Full path to the PFX file with the certificate
        :param password: Plain password used to decrypt the PFX file
        :param location: Where are you signing from i.e. "Steve's House, Cambridge"
        :param reason: Why are you signing i.e. "Fake University Degree Certificate"
        """
        log.debug('Loading PFX file: "{}"'.format(pfx_path))

        with open(pfx_path, 'rb') as cert_file:
            pfx = OpenSSL.crypto.load_pkcs12(cert_file.read(), password)
        
        self.private_key = pfx.get_privatekey().to_cryptography_key()
        cert = pfx.get_certificate()
        self.cert = cert.to_cryptography()

        name_parts = ['{}={}'.format(name.decode(), value.decode()) for name, value in cert.get_subject().get_components()]
        contact = ', '.join(name_parts)
        
        issuer_parts = ['{}={}'.format(name.decode(), value.decode()) for name, value in cert.get_issuer().get_components()]
        issuer = ', '.join(issuer_parts)

        serial_number = cert.get_serial_number()

        self.signature = Signature(
            contact=contact,
            issuer=issuer,
            serial_number=serial_number,
            location=location,
            reason=reason
        )

    def get_pdf_signature(self, unsigned_pdf):
        """
        :param unsigned_pdf: PDF data (bytes)
        :return: Signature to append to PDF (bytes)
        """
        return endesive.pdf.cms.sign(
            unsigned_pdf, self.signature.get_signature_args(), self.private_key, self.cert, [], 'sha256'
        )

    def get_signed_pdf(self, unsigned_pdf):
        """
        :param unsigned_pdf: PDF data (bytes)
        :return: PDF file with signature attached
        """
        return unsigned_pdf + self.get_pdf_signature(unsigned_pdf)



"""
Contains a queue of email messages to be sent asynchronously
"""

import logging
import datetime
import traceback

from . import systemevent
from .. import timetool
from .. import lfsmailer

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

STATUS_QUEUED = 'QUEUED'
STATUS_SENDING = 'SENDING'
STATUS_SENT = 'SENT'
STATUS_FAILED = 'FAILED'
STATUS_RETRY = 'RETRY'

_db = None
_QueuedEmail = None

# QueuedEmail model:
#
# class QueuedEmail(db.Model):
#     id = db.Column(db.BigInteger, primary_key=True, nullable=False)
#     html = db.Column(db.Boolean, nullable=False)
#     timestamp = db.Column(db.DateTime, nullable=False)
#     to_addresses = db.Column(db.String, nullable=False)
#     from_address = db.Column(db.String, nullable=False)
#     subject = db.Column(db.String, nullable=False)
#     body = db.Column(db.String, nullable=False)
#     status = db.Column(db.String, nullable=False)
#     attempts = db.Column(db.SmallInteger, nullable=False)
#     last_send_timestamp = db.Column(db.DateTime, nullable=True)
#
#     def __init__(self, html, to_addresses, from_address, subject, body, status):
#         self.html = html
#         self.to_addresses = to_addresses
#         self.from_address = from_address
#         self.subject = subject
#         self.body = body
#
#         self.timestamp = datetime.datetime.utcnow()
#         self.status = status
#         self.attempts = 0


def init(db, queued_email_model_class):
    """
    :param db: Flask-SQLAlchemy database object
    :param queued_email_model_class: Model class (see above example)
    """
    global _db, _QueuedEmail

    _db = db
    _QueuedEmail = queued_email_model_class


class ProcessEmailQueueSystemEvent(systemevent.SystemEvent):
    def __init__(self, app):
        run_every = app.config['PROCESS_EMAIL_QUEUE_INTERVAL']
        first_run = datetime.datetime.utcnow() + datetime.timedelta(seconds=run_every)

        super(ProcessEmailQueueSystemEvent, self).__init__(run_every=run_every, first_run=first_run)

        self.max_attempts = app.config['EMAIL_QUEUE_MAX_ATTEMPTS']

    def process(self):
        queued_emails = _QueuedEmail.query.filter(_db.or_(_QueuedEmail.status == STATUS_QUEUED,
                                                          _QueuedEmail.status == STATUS_RETRY)).all()
        for email in queued_emails:
            email.status = STATUS_SENDING
            email.attempts += 1
            _db.session.commit()
            try:
                recipients = [s.strip() for s in email.to_addresses.split(',')]
                lfsmailer.send_mail(recipients, email.subject, email.body, html=email.html,
                                    from_address=email.from_address)
                email.status = STATUS_SENT
                _db.session.commit()
            except Exception as e:
                if email.attempts < self.max_attempts:
                    stack_trace = traceback.format_exc()

                    log.warning('Email attempt failed, retrying. Exception: %s\n%s' % (e, stack_trace))
                    email.status = STATUS_RETRY
                    _db.session.commit()
                else:
                    log.exception('Email sending failed. Max retries exceeded')
                    email.status = STATUS_FAILED
                    _db.session.commit()


class EmailQueueCleanUpSystemEvent(systemevent.SystemEvent):
    def __init__(self, app):
        super(EmailQueueCleanUpSystemEvent, self).__init__(run_every=app.config['CLEAN_UP_EMAIL_QUEUE_INTERVAL'])

        self.max_age = app.config['EMAIL_QUEUE_MAX_DAYS']

    def process(self):
        max_age = datetime.timedelta(days=self.max_age)
        delete_time = datetime.datetime.utcnow() - max_age
        log.info('Cleaning up emails queued before %s' % timetool.format_datetime_long(delete_time))
        num_deleted = _QueuedEmail.query.filter(_QueuedEmail.timestamp < delete_time).delete()
        _db.session.commit()
        log.info('Deleted %s emails' % num_deleted)


def queue_email(to_addresses, from_address, subject, body, commit=True, html=True, session=None):
    """
    Add a mail to the queue to be sent.

    WARNING: Commits by default!

    :param to_addresses: The names and addresses to send the email to, i.e. "Steve<steve@fig14.com>, info@fig14.com"
    :param from_address: Who the email is from i.e. "Stephen Brown <s@fig14.com>"
    :param subject: The email subject
    :param body: The html / text body of the email
    :param commit: Whether to commit to the database
    :param html: Is this a html email?
    :param session: The sqlalchemy session or None to use db.session
    """
    from models import QueuedEmail

    if session is None:
        session = _db.session

    log.info('Queuing mail to %s: %s' % (to_addresses, subject))
    queued_email = QueuedEmail(html, to_addresses, from_address, subject, body, STATUS_QUEUED)
    session.add(queued_email)
    session.commit()

    return queued_email


def queue_html_email(to_addresses, from_address, subject, body, commit=True, session=None):
    return queue_email(to_addresses, from_address, subject, body, commit=commit, html=True, session=session)


def queue_text_email(to_addresses, from_address, subject, body, commit=True, session=None):
    return queue_email(to_addresses, from_address, subject, body, commit=commit, html=False, session=session)




"""
Functions for saving warning and errors to the database
"""

import logging
import io
import traceback
import datetime

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, BigInteger, String, DateTime, func
import sqlalchemy.exc


__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

_ERROR_REPORT_THRESHOLD = None

# Database connection
_connection = None

# Session maker
_Session = None

# Database Model
WarningLogMessage = None

# Celery Task
save_warninglog_message_task = None


def init(app, engine_or_connection, celery=None, table_name='warning_log_message'):
    global _ERROR_REPORT_THRESHOLD, _connection, WarningLogMessage, \
        _Session, save_warninglog_message_task
    
    _ERROR_REPORT_THRESHOLD = app.config['WARNINGLOG_ERROR_REPORT_THRESHOLD']
    _connection = engine_or_connection

    BaseModel = declarative_base(bind=_connection)
    _Session = sessionmaker(bind=_connection)

    session = _Session()

    class DatabaseModel(BaseModel):
        __tablename__ = table_name

        id = Column(BigInteger, primary_key=True, nullable=False)
        timestamp = Column(DateTime, nullable=False)
        log_name = Column(String, nullable=False)
        location = Column(String, nullable=False)
        function = Column(String, nullable=True)
        title = Column(String, nullable=False)
        description = Column(String, nullable=False)
        traceback = Column(String, nullable=True)

        def __init__(self, timestamp, log_name, location, function, title, description, traceback=None):
            """
            :param timestamp: The time that the warning was raised
            :param log_name: Name of module (__name__ from python file)
            :param location: Full path and line number, i.e. "littlefish.timetool.py:113"
            :param function: The name of the function, i.e. "format_datetime"
            :param title: Error title (used to deduplicate)
            :param description: Full error string
            :param traceback: Optional string with traceback
            """
            self.timestamp = datetime.datetime.utcnow()
            self.log_name = log_name
            self.location = location
            self.function = function
            self.title = title
            self.description = description
            self.traceback = traceback

    WarningLogMessage = DatabaseModel

    # Create the table if it's missing
    try:
        WarningLogMessage.__table__.create()
    except sqlalchemy.exc.ProgrammingError:
        # Probably already exists so do nothing
        pass

    # Check that the table actually exists by trying to select from it
    try:
        session.query(WarningLogMessage).first()
    except Exception as e:
        raise Exception('Could not query from {}. Table creation probably failed!'.format(table_name), e)
    
    if celery:
        log.info('Using Celery for warninglog')
        save_warninglog_message_task = celery.task(save_warninglog_message)
    else:
        log.info('Not using Celery for warninglog - warnings will be logged in main thread')


def save_warninglog_message(timestamp, log_name, location, function, title, description, traceback=None):
    """
    :param log_name: Name of module (__name__ from python file)
    :param location: Full path and line number, i.e. "littlefish.timetool.py:113"
    :param function: The name of the function, i.e. "format_datetime"
    :param title: Error title (used to deduplicate)
    :param description: Full error string
    :param traceback: Optional string with traceback
    """
    log.info('Warning Log: {} - {}'.format(title, description))
    message = WarningLogMessage(timestamp, log_name, location, function, title, description, traceback)
    session = _Session()
    session.add(message)

    try:
        session.commit()
    except Exception:
        session.rollback()
        log.exception('Error saving warning log message to database: {}'.format(message.strip()))

    session.close()


def _find_caller(stack_info=False):
    """
    Shamelessly stolen from the logging library

    Find the stack frame of the caller so that we can note the source
    file name, line number and function name.
    """
    f = logging.currentframe()
    
    rv = "(unknown file)", 0, "(unknown function)", None
    while hasattr(f, "f_code"):
        co = f.f_code
        sinfo = None
        if stack_info:
            sio = io.StringIO()
            sio.write('Stack (most recent call last):\n')
            traceback.print_stack(f, file=sio)
            sinfo = sio.getvalue()
            if sinfo[-1] == '\n':
                sinfo = sinfo[:-1]
            sio.close()
        rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
        break
    return rv


class WarningLogSummary(object):
    """
    Used in the summary emails
    """
    def __init__(self, title, count):
        self.title = title
        self.count = count


class WarningLog(object):
    """
    Usage:

    from littlefish.warninglog import WarningLog

    warning_log = WarningLog(__name__)

    def some_function():
        # Log a simple message
        warning_log.log_message('This is a warning!')

        try:
            a = 4 / 0
        except:
            # Log an exception
            warning_log.log_exception('An exception was raised')
    """
    def __init__(self, name):
        self.name = name
    
    def _log(self, location, function, title, description, traceback_string=None):
        """
        Log a message to the WarningLogMessage table
        
        :param location: The filename and number
        :param function: The function name
        :param title: The title - should be the same for all similar errors
        :param description:  Detailed description of the error
        :param traceback_string: String with traceback (optional)
        """
        if WarningLogMessage is None:
            raise Exception('Tried to log a warning with warninglog without '
                            'calling init()')

        message = '{}:{} - {}'.format(self.name, title, description)
        if traceback_string:
            message = '{}\n{}'.format(message, traceback_string)
        
        # If the error has been reported more tha ERROR_REPORT_THRESHOLD today, we need
        # to log an error, otherwise just a warning
        today = datetime.date.today()
        period_start = datetime.datetime(today.year, today.month, today.day)
        
        log_warning = True

        if _ERROR_REPORT_THRESHOLD is not None:
            number_of_errors = 0

            try:
                # Create another session so that we can query from the database even if it's in an inconsistent state
                session = _Session()

                number_of_errors = session.query(WarningLogMessage).filter(
                    WarningLogMessage.title == title,
                    WarningLogMessage.timestamp >= period_start
                ).count()

                session.close()

            except Exception:
                log.exception('Couldn\'t load error count when logging warning')
                number_of_errors = 0

            log_warning = number_of_errors < _ERROR_REPORT_THRESHOLD
        
        if log_warning:
            log.warning(message)
        else:
            log.error('Warning log message has been reported {}/{} times today: {}'.format(
                number_of_errors, _ERROR_REPORT_THRESHOLD, message
            ))
        
        if save_warninglog_message_task:
            # Use Celery if possible
            save_warninglog_message_task.delay(datetime.datetime.utcnow(), self.name, location, function, title,
                                               description, traceback_string)
        else:
            save_warninglog_message(datetime.datetime.utcnow(), self.name, location, function, title,
                                    description, traceback_string)
    
    def log_message(self, title, description='(no further details)'):
        """
        Log a message to the WarningLog table
        
        :param title: The title - should be the same for all similar errors
        :param description:  Detailed description of the error
        """
        file_name, line_number, func, stack_info = _find_caller()
        
        try:
            location = '{}:{}'.format(file_name, line_number)
        except Exception:
            location = 'UNKNOWN:???'
        
        self._log(location, func, title, description)
        
    def log_exception(self, title, description='(no further details)'):
        """
        Log a message with the current traceback
        
        :param title: The title - should be the same for all similar errors
        :param description:  Detailed description of the error
        """
        file_name, line_number, func, stack_info = _find_caller()
        
        f = io.StringIO()
        traceback.print_exc(file=f)
        traceback_string = f.getvalue()
        
        try:
            location = '{}:{}'.format(file_name, line_number)
        except Exception:
            location = 'UNKNOWN:???'
            
        self._log(location, func, title, description, traceback_string)


def get_warninglog_summary(start_timestamp=None, end_timestamp=None):
    """
    If no timestamps are passed in this will return yesterday's data
    """
    now = datetime.datetime.utcnow()
    today = datetime.datetime(year=now.year, month=now.month, day=now.day)
    yesterday = today - datetime.timedelta(days=1)

    if start_timestamp is None:
        start_timestamp = yesterday

    if end_timestamp is None:
        end_timestamp = today

    if end_timestamp <= start_timestamp:
        raise ValueError('Invalid date range: {} to {}'.format(start_timestamp, end_timestamp))
    
    # SELECT title, COUNT(*) FROM warning_log_message GROUP BY title;
    session = _Session()
    rows = session.query(
        WarningLogMessage.title, func.count(WarningLogMessage.id)
    ).filter(
        WarningLogMessage.timestamp >= yesterday,
        WarningLogMessage.timestamp < today
    ).group_by(
        WarningLogMessage.title
    ).order_by(
        WarningLogMessage.title
    ).all()

    summary = []
    for row in rows:
        summary.append(WarningLogSummary(row[0], row[1]))

    return summary



"""
Base code for building site session data classes - classes which manage the session dict, and
handle serialisation of data into session compatible types
"""

import logging

from flask import session

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class _NoDefaultType(object):
    pass


# Set default value to this to avoid setting value to None by default
NoDefault = _NoDefaultType()


class SessionVar(object):
    def __init__(self, key, default_value=None):
        self.key = key
        self.default_value = default_value
    
    def __get__(self, instance, owner):
        if instance is None:
            raise Exception('SessionVars can only be set through SessionData instances')

        if instance._has_session_value(self.key):
            json_val = instance._get_session_value(self.key)
            if json_val is None:
                return None
            return self.from_json_val(json_val)

        elif self.default_value == NoDefault:
            raise AttributeError('Attribute "{}" hasn\'t been set for {} instance'.format(
                self.key, owner.__name__
            ))
        else:
            if isinstance(self.default_value, list):
                # Copy list to avoid appending to actual default value
                default_value = self.default_value[:]
            elif isinstance(self.default_value, dict):
                # Copy dict
                default_value = self.default_value.copy()
            else:
                # Just return the default value
                default_value = self.default_value

            return default_value

    def __set__(self, instance, value):
        if instance is None:
            raise Exception('SessionVars can only be set through SessionData instances')
        
        if value is None:
            instance._set_session_value(self.key, None)
        else:
            instance._set_session_value(self.key, self.to_json_val(value))

    def to_json_val(self, val):
        """
        Convert from the datatype of this SessionVar back into a json compatible value.

        This is the method you override to implement the desired type conversion when saving to
        the session
        """
        return val

    def from_json_val(self, json_val):
        """
        Convert from a json compatible value, back into the correct type for this SessionVar.

        This is the method you override to implement the desired type conversion when loading from
        the session
        """
        return json_val


class SessionData(object):
    def __init__(self, version=1, version_key='_v', base_key=None):
        self._version = version
        self._version_key = version_key

        if base_key is None:
            base_key = self.__class__.__name__

        self._base_key = base_key

        if base_key not in session:
            session[self._base_key] = {}

        self._data_dict = session[base_key]
        
        session_version = self._data_dict.get(version_key, -1)
        if session_version < self._version:
            if self._version > -1:
                log.info('Clearing old session (version = {}, latest version = {}'.format(
                    session_version, self._version
                ))
            self.clear()
        
        # Ensure session is always saved
        session.modified = True

    def clear(self):
        self._data_dict.clear()
        self._data_dict[self._version_key] = self._version
    
    def _has_session_value(self, key):
        return key in self._data_dict

    def _get_session_value(self, key):
        return self._data_dict[key]

    def _set_session_value(self, key, value):
        self._data_dict[key] = value

    def get_data(self):
        """
        :return: the raw data dictionary. This can be used to restore the sessiondata after the
                 flask session is cleared
        """
        return self._data_dict

    def restore_data(self, data_dict):
        """
        Restore the data dict - update the flask session and this object
        """
        session[self._base_key] = data_dict
        self._data_dict = session[self._base_key]




"""
Unit tests for util.py
"""

import pytest

from littlefish.util import format_number_1dp, format_number_2dp, format_price_commas_no_point,\
    format_commas, format_number_2dp_trailing_zeros, format_price, format_multiline_html,\
    to_paragraphs, format_ordinal, format_filesize, is_ascii, extension_from_filename

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'


def test_format_number_1dp():
    assert format_number_1dp(1) == '1'
    assert format_number_1dp(101) == '101'
    assert format_number_1dp(101.1) == '101.1'
    assert format_number_1dp(101.01) == '101'
    assert format_number_1dp(101.001) == '101'
    assert format_number_1dp(101.162451532344) == '101.2'
    assert format_number_1dp(-234.56789) == '-234.6'
    assert format_number_1dp(None) is None


def test_format_number_2dp():
    assert format_number_2dp(1) == '1'
    assert format_number_2dp(101) == '101'
    assert format_number_2dp(101.1) == '101.1'
    assert format_number_2dp(101.01) == '101.01'
    assert format_number_2dp(101.001) == '101'
    assert format_number_2dp(101.162451532344) == '101.16'
    assert format_number_2dp(-234.56789) == '-234.57'
    assert format_number_2dp(None) is None


def test_format_price_commas_no_point():
    def f(x):
        return str(format_price_commas_no_point(x))

    assert f(1) == '&pound;1'
    assert f(2.00) == '&pound;2'
    assert f(0.1) == '&pound;0'
    assert f(123456.78) == '&pound;123,457'
    assert f(123456789.01) == '&pound;123,456,789'
    assert format_price_commas_no_point(None) is None


def test_format_commas():
    def f(x):
        return str(format_commas(x))

    assert f(1) == '1'
    assert f(2.00) == '2'
    assert f(0.1) == '0'
    assert f(123456.78) == '123,457'
    assert f(123456789.01) == '123,456,789'
    assert format_commas(None) is None


def test_format_number_2dp_trailing_zeros():
    assert format_number_2dp_trailing_zeros(1) == '1.00'
    assert format_number_2dp_trailing_zeros(10) == '10.00'
    assert format_number_2dp_trailing_zeros(10.1) == '10.10'
    assert format_number_2dp_trailing_zeros(10.11) == '10.11'
    assert format_number_2dp_trailing_zeros(10.111) == '10.11'
    assert format_number_2dp_trailing_zeros(10.119) == '10.12'
    assert format_number_2dp_trailing_zeros(10.116158426456) == '10.12'
    assert format_number_2dp_trailing_zeros(0.6) == '0.60'
    assert format_number_2dp_trailing_zeros(-10.11615) == '-10.12'
    assert format_number_2dp_trailing_zeros(None) is None


def test_format_price():
    def f(x):
        return str(format_price(x))

    assert f(1) == '&pound;1.00'
    assert f(2.00) == '&pound;2.00'
    assert f(0.1) == '&pound;0.10'
    assert f(123456.78) == '&pound;123456.78'
    assert f(123456789.01) == '&pound;123456789.01'
    assert f(123456789.01698) == '&pound;123456789.02'
    assert format_price(None) is None


def test_format_multiline_html():
    def f(x):
        result = format_multiline_html(x)
        return str(result) if result else None

    assert f(None) is None
    assert f('test sentence') == 'test sentence'
    assert f('test sentence\rreturn') == 'test sentencereturn'
    assert f('test sentence\nreturn') == 'test sentence<br>return'
    assert f('test sentence\r\nreturn') == 'test sentence<br>return'
    assert f('test sentence\n\rreturn') == 'test sentence<br>return'
    assert f('test \rsentence\nreturn') == 'test sentence<br>return'
    assert f('test sentence\r\nreturn\n\rreturn') == 'test sentence<br>return<br>return'


def test_to_paragraphs():
    def f(x):
        return str(to_paragraphs(x))

    assert f('test') == '<p>test</p>'
    assert f('a longer test sentence') == '<p>a longer test sentence</p>'
    assert f('a\ntest sentence') == '<p>a\ntest sentence</p>'
    assert f('a\n\ntest sentence') == '<p>a</p><p>test sentence</p>'
    assert f('a\n\ntest\n\nsentence') == '<p>a</p><p>test</p><p>sentence</p>'
    assert f('a\n\n\ntest\n\n\n\nsentence') == '<p>a</p><p>test</p><p>sentence</p>'


def test_format_ordinal():
    assert format_ordinal(0) == '0th'
    assert format_ordinal(1) == '1st'
    assert format_ordinal(2) == '2nd'
    assert format_ordinal(3) == '3rd'
    assert format_ordinal(4) == '4th'
    assert format_ordinal(10) == '10th'
    assert format_ordinal(11) == '11th'
    assert format_ordinal(12) == '12th'
    assert format_ordinal(21) == '21st'
    assert format_ordinal(22) == '22nd'
    assert format_ordinal(23) == '23rd'
    assert format_ordinal(400256) == '400256th'
    assert format_ordinal(13451263) == '13451263rd'
    assert format_ordinal(4156426442) == '4156426442nd'
    assert format_ordinal(98765432111111) == '98765432111111th'
    assert format_ordinal(98765432111101) == '98765432111101st'

    with pytest.raises(TypeError):
        format_ordinal(None)


def test_format_filesize():
    assert format_filesize(0) == '0 Bytes'
    assert format_filesize(1) == '1 Bytes'
    assert format_filesize(492) == '492 Bytes'
    assert format_filesize(1023) == '1023 Bytes'
    assert format_filesize(1024) == '1.0 KiB'
    assert format_filesize(452687) == '442.1 KiB'
    assert format_filesize(1048576) == '1.0 MiB'
    assert format_filesize(8152324564515) == '7774662.6 MiB'


def test_is_ascii():
    assert is_ascii(b'abcdefghijklmnopqrstuvwxyz') is True
    assert is_ascii(b'abcdefghijklmnop\xc2\xa3') is False


def test_extension_from_filename():
    assert extension_from_filename('testing.jpg') == 'jpg'
    assert extension_from_filename('testing123.png') == 'png'
    assert extension_from_filename('testing-again.pdf') == 'pdf'
    assert extension_from_filename('testing.test.txt') == 'txt'



"""
Utility functions for processing HTML
"""

import logging
import re

from bs4 import BeautifulSoup

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

INLINE_TAGS = ['a', 'span', 'em', 'strong', 'b', 'big', 'i', 'small', 'tt', 'abbr', 'acronym',
               'cite', 'code', 'em', 'img', 'sub', 'sup']


def get_all_text(html):
    soup = BeautifulSoup(html, "lxml")
    return soup.find_all(text=True)


def tag_to_text(tag):
    """
    :param tag: Beautiful soup tag
    :return: Flattened text
    """
    out = []
    for item in tag.contents:
        # If it has a name, it is a tag
        if item.name:
            out.append(tag_to_text(item))
        else:
            # Just text!
            out.append(item)

    return ' '.join(out)


def split_line(line, min_line_length=30, max_line_length=100):
    """
    This is designed to work with prettified output from Beautiful Soup which indents with a single space.

    :param line: The line to split
    :param min_line_length: The minimum desired line length
    :param max_line_length: The maximum desired line length

    :return: A list of lines
    """
    if len(line) <= max_line_length:
        # No need to split!
        return [line]

    # First work out the indentation on the beginning of the line
    indent = 0
    while line[indent] == ' ' and indent < len(line):
        indent += 1

    # Try to split the line
    # Start looking for a space at character max_line_length working backwards
    i = max_line_length
    split_point = None
    while i > min_line_length:
        if line[i] == ' ':
            split_point = i
            break
        i -= 1

    if split_point is None:
        # We didn't find a split point - search beyond the end of the line
        i = max_line_length + 1
        while i < len(line):
            if line[i] == ' ':
                split_point = i
                break
            i += 1

    if split_point is None:
        # There is nowhere to split the line!
        return [line]
    else:
        # Split it!
        line1 = line[:split_point]
        line2 = ' ' * indent + line[split_point + 1:]
        return [line1] + split_line(line2, min_line_length, max_line_length)


def pretty_print(html, max_line_length=110, tab_width=4):
    """
    Pretty print HTML, splitting it into lines of a reasonable length (if possible).

    This probably needs a whole lot more testing!

    :param html: The HTML to format
    :param max_line_length: The desired maximum line length. Will not be strictly adhered to
    :param min_line_length: The desired minimum line length
    :param tab_width: Essentially, the tabs that indent the code will be treated as this many
                      spaces when counting the length of each line

    :return: Beautifully formatted HTML
    """
    if tab_width < 2:
        raise ValueError('tab_width must be at least 2 (or bad things would happen!)')

    # Double curly brackets to avoid problems with .format()
    html = html.replace('{', '{{').replace('}', '}}')
    soup = BeautifulSoup(html, 'lxml')
    soup.html.unwrap()
    soup.body.unwrap()

    unformatted_tag_list = []
    
    # Here we are taking the tags out of the content and replacing them with placeholders,
    # and adding the tags to a list.  I didn't come up with this...
    for i, tag in enumerate(soup.find_all(INLINE_TAGS)):
        unformatted_tag_list.append(str(tag))
        tag.replace_with('{' + 'unformatted_tag_list[{0}]'.format(i) + '}')

    # If we prettify this, there will still be some weird indentation going on, based on
    # the original markup, so we need to convert it into a string again, and then parse
    # it again
    processed_html = str(soup)
    soup2 = BeautifulSoup(processed_html, 'lxml')
    soup2.html.unwrap()
    soup2.body.unwrap()

    # Prettify it, substitute in the unformatted tags
    pretty_markup = soup2.prettify().format(unformatted_tag_list=unformatted_tag_list)

    # Convert indendtations to a tab width of 4
    pretty_markup = re.sub(r'^(\s+)', r'\1' * tab_width, pretty_markup, flags=re.MULTILINE)

    # Final step - pass over the formatted html, convert the indentations into tabs and cut
    # the lines to length
    lines = pretty_markup.splitlines()
    out = ''
    for line in lines:
        for line_part in split_line(line, max_line_length=max_line_length):
            out += line_part
            out += '\n'
    
    # Final final step! Convert space indentations into tabs
    return out.replace(' ' * tab_width, '\t')



"""
Shamelessly stolen from stack overflow.  Maps integer values onto
strings of characters - useful for generating things like voucher codes
"""

import logging

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class Encoder:
    def __init__(self, alphabet):
        """
        Creates an encoder which maps integers onto characters from
        the string alphabet.  Similar to base64 encoding but with a
        custom set of characters

        :param alphabet: The characters to use to encode the numeric
                         values, i.e.:
                         "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        """
        self.alphabet = alphabet
        self.size = len(alphabet)
        # Map from char -> value
        self.mapping = dict((ch, i) for (i, ch) in enumerate(alphabet))
        if len(self.mapping) != self.size:
            raise Exception("Duplicate characters in alphabet string")

    def encode(self, x):
        # Only needed if don't want '' for 0
        if x == 0:
            return self.alphabet[0]
        
        out = []
        while x > 0:
            out.append(self.alphabet[x % self.size])
            x //= self.size
        return ''.join(out)

    def decode(self, s):
        return sum(self.mapping[ch] * self.size ** i for (i, ch) in enumerate(s))



"""
Utility Functions
"""

import logging
import os
from io import StringIO
import csv
from functools import wraps
import random
import re

from flask import jsonify, make_response, current_app, request
import flask
from flask_sqlalchemy import get_debug_queries
import jinja2

# For backwards compatibility
from .fileutil import extension_from_filename, read_file, write_file  # noqa

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

LOWER_CASE = 'abcdefghijkmnopqrstuvwxyz'
UPPER_CASE = 'ABCDEFGHJKLMNOPQRSTUVWXYZ'
NUMBERS = '0123456789'
SYMBOLS = '!?@%$'

BASE62_MAP = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

PHONE_NUMBER_DISALLOWED_SYMBOLS = re.compile('[^0-9+]+')

paragraph_split_regex = re.compile(r'(\r\n|\n\r|\n){2,}')


class UpdateLogger(object):
    """
    Usefull class for logging updates to things when you're importing them
    """
    def __init__(self, prefix='', use_logger=False):
        self.prefix = prefix
        self.lines = []
        self.use_logger = use_logger
    
    def log_update(self, s):
        self.lines.append('{}{}'.format(self.prefix, s))

    def log_variable_update(self, variable_name, old_value=None, new_value=None):
        if old_value is None and new_value is None:
            self.lines.append('{}Updating {}'.format(
                self.prefix, variable_name
            ))
        else:
            self.lines.append('{}Updating {} from {} -> {}'.format(
                self.prefix, variable_name, old_value, new_value
            ))

    def log_item_added(self, item_type, value):
        self.lines.append('{}+ Adding {}: {}'.format(
            self.prefix, item_type, value
        ))
    
    def log_item_deleted(self, item_type, value):
        self.lines.append('{}- Deleting {}: {}'.format(
            self.prefix, item_type, value
        ))

    def log_item_modified(self, item_type, value):
        self.lines.append('{}~ Modifying {}: {}'.format(
            self.prefix, item_type, value
        ))

    @property
    def has_updates(self):
        return len(self.lines) > 0

    def print_updates(self, clear=True):
        if self.has_updates:
            for line in self.lines:
                if self.use_logger:
                    log.info(line)
                else:
                    print(line)

            if clear:
                self.lines = []


def generate_password():
    return random.choice(UPPER_CASE)\
        + random.choice(LOWER_CASE)\
        + random.choice(LOWER_CASE)\
        + random.choice(LOWER_CASE)\
        + random.choice(LOWER_CASE)\
        + random.choice(NUMBERS)\
        + random.choice(NUMBERS)\
        + random.choice(SYMBOLS)


def generate_secure_password(length):
    if length < 6:
        raise ValueError('Length must be at least 6')

    num_symbols = 2
    num_numbers = 1
    num_other = length - num_symbols - num_numbers

    characters = []
    for i in range(num_symbols):
        characters.append(random.choice(SYMBOLS))

    for i in range(num_numbers):
        characters.append(random.choice(NUMBERS))
    
    # Note: we can have more numbers
    other_characters = LOWER_CASE + UPPER_CASE + NUMBERS
    for i in range(num_other):
        characters.append(random.choice(other_characters))

    # Shuffle the characters
    random.shuffle(characters)

    return ''.join(characters)


def ajax_error(error_message):
    return jsonify({'status': 'error', 'message': error_message})


def format_number_1dp(number):
    if number is None:
        return None
    return ("%.1f" % number).rstrip('0').rstrip('.')


def format_number_2dp(number):
    if number is None:
        return None
    return ("%.2f" % number).rstrip('0').rstrip('.')


def format_number_2dp_trailing_zeros(number):
    if number is None:
        return None
    return "%.2f" % number


def format_number_2_to_4_dp(number):
    if number is None:
        return None

    formatted = '{0:.4f}'.format(number)
    return re.sub(r'(.*\.[0-9][0-9][1-9]*)0*$', r'\1', formatted)


def format_price(price):
    if price is None:
        return None
    if price >= 0:
        return jinja2.Markup('&pound;%0.2f' % price)
    else:
        return jinja2.Markup('-&pound;%0.2f' % -price)


def format_price_commas(price):
    """
    Formats prices, rounding (i.e. to the nearest whole number of pounds) with commas
    """
    if price is None:
        return None
    if price >= 0:
        return jinja2.Markup('&pound;{:,.2f}'.format(price))
    else:
        return jinja2.Markup('-&pound;{:,.2f}'.format(-price))


def format_price_commas_no_point(price):
    """
    Formats prices, rounding (i.e. to the nearest whole number of pounds) with commas
    """
    if price is None:
        return None
    return jinja2.Markup('&pound;{:,.0f}'.format(price))


def format_commas(number):
    """
    Rounds a number and formats it with commas i.e. 123,456,789
    """
    if number is None:
        return None
    return '{:,.0f}'.format(number)


def format_multiline_html(text):
    """
    Turns a string like 'a\nb\nc' into 'a<br>b<br>c' and marks as Markup

    Note: Will remove all \r characters from output (if present)
    """
    if text is None:
        return None

    if '\n' not in text:
        return text.replace('\r', '')

    parts = text.replace('\r', '').split('\n')
    out = flask.Markup()
    for part in parts:
        if out:
            out += flask.Markup('<br>')
        out += flask.escape(part)
    return out


def to_paragraphs(string):
    # Split the string into parts using multiple returns as a separator
    parts = paragraph_split_regex.split(string)
    return flask.Markup('<p>%s</p>' % '</p><p>'.join(parts))


def format_filesize(bytes):
    one_mb = 1024 * 1024

    if bytes < 1024:
        return '%s Bytes' % bytes
    elif bytes < one_mb:
        return '%0.1f KiB' % (float(bytes) / 1024)
    else:
        return '%0.1f MiB' % (float(bytes) / one_mb)


def format_ordinal(n):
    """
    Format an ordinal, like 1st, 2nd, 3rd...

    Not tested with large numbers of negative numbers!
    """
    if 10 <= n % 100 < 20:
        return str(n) + 'th'
    else:
        return str(n) + {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, "th")


def format_phone_number(phone_number):
    """Removes unnecessary symbols from phone numbers"""
    if phone_number is None:
        return None

    return PHONE_NUMBER_DISALLOWED_SYMBOLS.sub('', phone_number)


def is_ascii(s):
    """
    Check if a "bytes" contains all ascii characters
    :param s: the string to check
    :return: true if the string is entirely ascii characters
    """
    try:
        s.decode('ascii')
    except UnicodeDecodeError:
        return False
    return True


def ensure_dir(path):
    """Ensure that a needed directory exists, creating it if it doesn't"""
    try:
        log.info('Ensuring directory exists: %s' % path)
        os.makedirs(path)
    except OSError:
        if not os.path.isdir(path):
            raise


def make_csv_response(csv_data, filename):
    """
    :param csv_data: A string with the contents of a csv file in it
    :param filename: The filename that the file will be saved as when it is downloaded by the user
    :return: The response to return to the web connection
    """
    resp = make_response(csv_data)
    resp.headers['Content-Type'] = 'application/octet-stream'
    resp.headers['Content-Disposition'] = 'attachment; filename=%s' % filename

    return resp


def rows_to_csv(rows, filename='report.csv', header=None):
    outfile = StringIO()
    writer = csv.writer(outfile)
    if header:
        writer.writerow(header)
    for row in rows:
        try:
            writer.writerow(row)
        except UnicodeEncodeError:
            replacement_row = []
            for item in row:
                # if item contains utf-8 characters, then replace string with 'INVALID'
                if not isinstance(item, str) or is_ascii(item):
                    replacement_row.append(item)
                else:
                    replacement_row.append('!!INVALID!!')
            writer.writerow(replacement_row)

    return make_csv_response(outfile.getvalue(), filename)


def no_cache(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        resp = make_response(f(*args, **kwargs))
        h = resp.headers
        h['Cache-Control'] = 'private, no-store, max-age=0, no-cache, must-revalidate, post-check=0, pre-check=0'
        h['Pragma'] = 'no-cache'
        h['Expires'] = 'Fri, 01 Jan 1990 00:00:00 GMT'
        return resp

    return decorated


def to_base62(n):
    """
    Converts a number to base 62
    :param n: The number to convert
    :return: Base 62 representation of number (string)
    """
    remainder = n % 62
    result = BASE62_MAP[remainder]
    num = n // 62

    while num > 0:
        remainder = num % 62
        result = '%s%s' % (BASE62_MAP[remainder], result)
        num = num // 62

    return result


def from_base62(s):
    """
    Convert a base62 String back into a number
    :param s: The base62 encoded String
    :return: The number encoded in the String (integer)
    """
    result = 0

    for c in s:
        if c not in BASE62_MAP:
            raise Exception('Invalid base64 string: %s' % s)

        result = result * 62 + BASE62_MAP.index(c)

    return result


def chunks(l, n):
    """Yield successive n-sized chunks from list"""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def generate_random_numeric_string(length):
    return ''.join(random.choice(NUMBERS) for x in range(length))


def deprecated(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        name = '{}.{}'.format(f.__module__, f.__name__)
        log.error('!!!!!!!!!!! DEPRECATED : {} !!!!!!!!!!!!!'.format(name))
        return f(*args, **kwargs)

    return decorated


def deprecated_class_method(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        self = args[0]
        name = '{}.{}.{}'.format(f.__module__, self.__class__.__name__, f.__name__)
        log.error('!!!!!!!!!!! DEPRECATED : {} !!!!!!!!!!!!!'.format(name))
        return f(*args, **kwargs)

    return decorated


def test_mode_only(f):
    """Decorator - marks a view as only being accessible when the server is started in test mode"""
    @wraps(f)
    def decorated(*args, **kwargs):
        # Check if the logged in user has a customer
        if not current_app.config['TEST_MODE']:
            return '<html><head><title>Disabled!</title></head><body>'\
                '<h1 style="color: red;">Disabled!</h1>'\
                '<p>This page is only accessible if the server is in TEST_MODE.<p>'\
                '</body></html>'

        # Let them view the page
        return f(*args, **kwargs)

    return decorated


def print_sql_debug_timings():
    try:
        log.info('***** SQL Profiling for request: %s *****' % request.url)
        if not current_app.config['SQL_TIMINGS_ONLY_SHOW_SUMMARY']:
            for query in get_debug_queries():
                log.info('QUERY: %s\nParameters: %s\nDuration: %fs\nContext: %s\n' % (query.statement, query.parameters,
                                                                                      query.duration, query.context))
        total_time = 0
        total_queries = 0
        for query in get_debug_queries():
            total_time += query.duration
            total_queries += 1

        log.info('<< %s queries took %fs >>' % (total_queries, total_time))
    except:
        log.exception('Exception in SQL profiling code.  Ignoring.')


def is_uk_mobile_number(phone_number):
    """Is the passed in phone number a uk mobile number"""
    stripped = phone_number.replace(' ', '')

    return stripped.startswith('+447') or stripped.startswith('07')




"""
Utilities to manage Celery
"""

import logging
from functools import wraps

from celery import current_task

from littlefish import lfsmailer
from littlefish import redisutil


__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class CeleryEmailHandler(lfsmailer.LfsSmtpHandler):
    def add_details(self, message):
        return message


def get_enable_celery_error_reporting_function(site_name, from_address):
    """
    Use this to enable error reporting.  You need to put the following in your tasks.py or wherever you
    want to create your celery instance:

    celery = Celery(__name__)

    enable_celery_email_logging = get_enable_celery_error_reporting_function('My Website [LIVE]', 'errors@mywebsite.com')
    after_setup_logger.connect(enable_celery_email_logging)
    after_setup_task_logger.connect(enable_celery_email_logging)
    """
    def enable_celery_email_logging(sender, signal, logger, loglevel, logfile, format, colorize, **kwargs):
        from celery import current_app
        log.info('>> Initialising Celery task error reporting for logger {}'.format(logger.name))
        
        send_errors = current_app.conf['CELERY_SEND_TASK_ERROR_EMAILS']
        send_warnings = current_app.conf['CELERY_SEND_TASK_WARNING_EMAILS']

        if send_errors or send_warnings:
            error_email_subject = '{} Celery ERROR!'.format(site_name)
            celery_handler = CeleryEmailHandler(from_address, current_app.conf['ADMIN_EMAILS'],
                                                error_email_subject)

            if send_warnings:
                celery_handler.setLevel(logging.WARNING)
            else:
                celery_handler.setLevel(logging.ERROR)
            
            logger.addHandler(celery_handler)

    return enable_celery_email_logging


def init_celery(app, celery):
    """
    Initialise Celery and set up logging

    :param app: Flask app
    :param celery: Celery instance
    """
    celery.conf.update(app.config)
    
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    
    return celery


def non_overlapping(blocking, key=None, timeout=None):
    """
    Celery decorator to make tasks "non-overlapping". When an instance of this task starts
    while another instance is already running:

    If blocking is set to True then the second task will wait for the first task's lock to be
    released.

    If blocking is set to False then the second task will abort.

    :param blocking: Whether or not to wait for the existing task to finish
    :param key: If set, this is the key used in Redis for locking. Otherwise the key is generated
                from the task name
    :param timeout: If set, this is how long the lock can be held for, in seconds, before being
                    automatically released
    """
    def decorator(f):
        @wraps(f)
        def outer(*args, **kwargs):
            base_key = key
            if base_key is None:
                base_key = current_task.name

            if not base_key:
                raise Exception('No key for non-overlapping task!')

            lock = redisutil.get_non_overlapping_task_lock(base_key, timeout)

            if blocking:
                # Blocking - will wait for the lock
                with lock:
                    return f(*args, **kwargs)
            else:
                # Non-blocking - will give up if the lock isn't free
                return_value = None
                raised_exception = None

                try:
                    have_lock = lock.acquire(blocking=False)
                    if have_lock:
                        return_value = f(*args, **kwargs)
                    else:
                        log.info('Aborting task - already running')
                except Exception as e:
                    raised_exception = e
                finally:
                    if have_lock:
                        lock.release()
                    
                    if raised_exception:
                        raise raised_exception

                    return return_value

        return outer
    return decorator


def non_overlapping_block(f):
    return non_overlapping(True)(f)


def non_overlapping_discard(f):
    return non_overlapping(False)(f)





import sys
from setuptools import setup

if sys.version_info.major < 3:
    sys.exit('Sorry, this library only supports Python 3')

setup(
    name='littlefish',
    packages=['littlefish', 'littlefish.background', 'littlefish.sessiondata'],
    include_package_data=True,
    version='0.0.48',
    description='Flask webapp utility functions by Little Fish Solutions LTD',
    author='Stephen Brown (Little Fish Solutions LTD)',
    author_email='opensource@littlefish.solutions',
    url='https://github.com/stevelittlefish/littlefish',
    download_url='https://github.com/stevelittlefish/littlefish/archive/v0.0.48.tar.gz',
    keywords=['flask', 'utility', 'time', 'pager'],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Framework :: Flask',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'geoip2>=2.4.2',
        'beautifulsoup4>=3.5.0.0',
        'Pillow>=4.0.0',
        'SQLAlchemy>=1.1.0',
        'PyMarkovChain>=1.8',
        'pytz>=2019.1',
        'python-dateutil>=2.6.0',
        'Flask>=0.12.0',
        'Flask-SQLAlchemy>=2.0',
        'Jinja2>=2.9.0',
        'lxml>=3.7.0',
        'IPy>=0.83',
        'redis>=3.2.1',
        'celery>=4.3.0',
        'endesive>=1.1.5',
        'pyOpenSSL>=19.0.0'
    ]
)




"""
Utilities for handling JSON
"""

import logging
import json
from decimal import Decimal

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class DecimalFloatEncoder(json.JSONEncoder):
    """
    Adds ability to encode Decimals as floats.  It's shit but sometimes useful!
    """
    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)

        return super().default(obj)



"""
Functions for interacting with Flask SQLAlchemy models
"""

import logging

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def fast_count(db, Model):  # noqa
    """
    Do a fast but sometimes inaccurate count (postgresql only).

    :param db: SQLAlchemy instance
    :param Model: Model class i.e. User, Order...
    """
    return db.session.execute(
        'SELECT n_live_tup FROM pg_stat_all_tables WHERE relname = :tablename',
        {'tablename': Model.__tablename__}
    ).scalar()



"""
Utilities to implement extra functionality using Redis
"""

import logging
import contextlib

import redis

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


connection = None
LOCK_TIMEOUT = None
GLOBAL_KEY_PREFIX = None

# Key templates for locks
LOCK_KEY_NON_OVERLAPPING_TASK = 'locks:non-overlapping:{base_key}'


@contextlib.contextmanager
def multi_key_lock(keys, timeout=None):
    if isinstance(keys, str):
        raise ValueError('You can\'t pass a single string into this function!')
    
    # Check for duplicates in list
    if len(keys) != len(set(keys)):
        raise ValueError('The passed in keyset contains duplicates: "{}"'.format(
            '", "'.join(keys)
        ))

    # Sort keys into alphabetical order
    keys = sorted(keys)
    
    if timeout is None:
        timeout = LOCK_TIMEOUT

    locks = []
    for key in keys:
        lock = connection.lock(format_key(key), timeout=timeout)
        locks.append(lock)
        
        log.debug('Acquiring lock "{}"'.format(lock.name))
        lock.acquire(blocking=True)
    
    log.debug('All locks acquired!')

    # Do whatever it is that needs doing!
    try:
        yield
    finally:
        # Now we release the locks in reverse order
        locks = reversed(locks)

        for lock in locks:
            log.debug('Releasing lock "{}"'.format(lock.name))
            lock.release()


def init(app):
    """
    Initialise this library.  The following config variables need to be in your Flask config:

    REDIS_HOST: The host of the Redis server
    REDIS_PORT: The port of the Redis server
    REDIS_PASSWORD: The password used to connect to Redis or None
    REDIS_GLOBAL_KEY_PREFIX: A short string unique to your application i.e. 'MYAPP'.  This will be turned into
                             a prefix like '~~MYAPP~~:' and will be used to allow multiple applications to share
                             a single redis server
    REDIS_LOCK_TIMEOUT: An integer with the number of seconds to wait before automatically releasing a lock.
                        A good number is 60 * 5 for 5 minutes.  This stops locks from being held indefinitely
                        if something goes wrong, but bear in mind this can also cause concurrency issues if
                        you ave a locking process that takes longer than this timeout!
    """
    global connection, LOCK_TIMEOUT, GLOBAL_KEY_PREFIX

    host = app.config['REDIS_HOST']
    port = app.config['REDIS_PORT']
    password = app.config['REDIS_PASSWORD']
    
    GLOBAL_KEY_PREFIX = '~~{}~~:'.format(app.config['REDIS_GLOBAL_KEY_PREFIX'])
    LOCK_TIMEOUT = app.config['REDIS_LOCK_TIMEOUT']

    connection = redis.StrictRedis(host=host, port=port, password=password)


def format_key(name):
    return '{}{}'.format(GLOBAL_KEY_PREFIX, name)


def get_lock(name, timeout=None):
    if timeout is None:
        timeout = LOCK_TIMEOUT

    return connection.lock(format_key(name), timeout=timeout)


def get_non_overlapping_task_lock(base_key, timeout=None):
    """
    Returns a lock for non-overlapping tasks

    :param base_key: The unique identifier, usually the task name
    """
    key = LOCK_KEY_NON_OVERLAPPING_TASK.format(base_key=base_key)
    return get_lock(key, timeout=timeout)





"""
This package contains code for managing session data, by creating SessionData classes, instances
of which map a subset of the session data onto attributes on an object.
"""

import logging

from .framework import SessionData, NoDefault  # noqa
from .sessionvars import *  # noqa

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)





"""
Debugging stuff
"""

import logging
import inspect
import pprint

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class DontPrint:
    pass


DONT_PRINT = DontPrint()


DEBUG_FORMAT = '@@@ {}'
LESSER_DEBUG_FORMAT = '  @ {}'


def lesser_debug_print(s):
    print(LESSER_DEBUG_FORMAT.format(s))


def debug_print(s=DONT_PRINT, **kwargs):
    if s is not DONT_PRINT:
        print(DEBUG_FORMAT.format(s))
    
    first = True
    for k, v in kwargs.items():
        if s is DONT_PRINT and first:
            debug_print('{} = {}'.format(k, v))
        else:
            lesser_debug_print('{} = {}'.format(k, v))

        first = False


def print_vars(**kwargs):
    for k, v in kwargs.items():
        pretty = pprint.pformat(v)
        debug_print('{}: {} = {}'.format(type(v), k, pretty))


def print_spacer(level=3):
    if level == 1:
        c = '#'
    elif level == 2:
        c = '='
    else:
        c = '-'

    print(c * 80)


def print_location(**kwargs):
    """
    :param kwargs: Pass in the arguments to the function and they will be printed too!
    """
    try:
        stack = inspect.stack()[1]
        debug_print('{}:{} {}()'.format(stack[1], stack[2], stack[3]))
    except Exception:
        debug_print('UNKNOWN LOCATION')

    for k, v in kwargs.items():
        lesser_debug_print('{} = {}'.format(k, v))



"""
This is a useful logger for debugging a specific problem, especially if you need to log
from multiple places (i.e. multiple instances of web app, celery workers) into the same
place, and your main log file is cluttered with junk!

This enables you to quickly and easily set up a log to the database, a file and stdout.
"""

import logging
import datetime

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, BigInteger, String, DateTime
from sqlalchemy.orm import sessionmaker
import sqlalchemy.exc

from . import hostidentifier

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

_debug_log = None

FILENAME = None
DISABLE_FILE_LOGGING = False
BIND = None
DISABLE_DB_LOGGING = False
IDENTIFIER = None


class LogDatabaseHandler(logging.Handler):
    """
    Logger to log to database using SQLAlchemy model
    """
    def __init__(self, bind, table_name):
        super().__init__()

        Model = declarative_base(bind=bind)
        self.Session = sessionmaker(bind=bind)

        session = self.Session()

        class DebugLog(Model):
            __tablename__ = table_name
            id = Column(BigInteger, primary_key=True, nullable=False)
            timestamp = Column(DateTime, nullable=False)
            log_name = Column(String, nullable=False)
            level = Column(String, nullable=False)
            message = Column(String, nullable=False)

        self.Model = DebugLog

        # Create the table if it's missing
        try:
            self.Model.__table__.create()
        except sqlalchemy.exc.ProgrammingError:
            # Probably already exists so do nothing
            pass

        # Check that the table actually exists by trying to select from it
        try:
            session.query(self.Model).first()
        except Exception as e:
            raise Exception('Could not query from {}. Table creation probably failed!'.format(table_name), e)
    
    def emit(self, record):
        try:
            session = self.Session()

            log_record = self.Model(
                timestamp=datetime.datetime.fromtimestamp(record.created),
                log_name=record.name,
                level=record.levelname,
                message=record.msg
            )

            session.add(log_record)
            session.commit()
        except Exception:
            # Fall back on system logger
            log = logging.getLogger(__name__)
            log.exception('Error saving log message in database: {}'.format(record.msg))


def init_file_logging(filename):
    global FILENAME, _debug_log

    assert not _debug_log
    FILENAME = filename


def init_database_logging(engine_or_connection):
    global BIND, _debug_log

    assert not _debug_log
    BIND = engine_or_connection


def get_logger():
    global _debug_log, IDENTIFIER

    if _debug_log is None:
        _debug_log = logging.getLogger('dbg-{}'.format(hostidentifier.identifier))
        _debug_log.setLevel(logging.DEBUG)
        _debug_log.propagate = False
        
        date_format = '%Y-%m-%d %H:%M:%S'

        # Log to file
        if FILENAME is not None and not DISABLE_FILE_LOGGING:
            file_handler = logging.FileHandler(FILENAME)
            file_handler.setLevel(logging.DEBUG)
            _debug_log.addHandler(file_handler)

            _debug_log.info('')
            _debug_log.info('@' * 100)
            _debug_log.info('')
            _debug_log.info('-' * 100)
            _debug_log.info('Debug Log started {}'.format(datetime.datetime.utcnow()))
            _debug_log.info('-' * 100)

            file_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s', date_format))
        
        # Log to DB
        if BIND is not None:
            db_handler = LogDatabaseHandler(BIND, '__debug_log')
            db_handler.setLevel(logging.INFO)
            _debug_log.addHandler(db_handler)

        # Log to stdout
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter('@@@ %(asctime)s %(message)s @@@', date_format))
        stream_handler.setLevel(logging.DEBUG)
        _debug_log.addHandler(stream_handler)

        # Done!
        _debug_log.info('Debug log configured with the following handlers: {}'.format(_debug_log.handlers))

    return _debug_log


